#!/usr/bin/env python3
import os
import readline
import socket
import subprocess
import sys
import tempfile
import time

# Local Imports
# telnetlib is removed from stdlib in Python 3.13. 
# We keep a copy from Python 3.12 to avoid disruption.
# TODO: investigate other ways to talk to Festival and other synths.
from my_telnetlib import Telnet

prompt = '"'

class Festival:
    "Festival Speech Synthesizer controlled via Pipe"
    x = None

    def comm(self, command):
        self.x.stdin.write(command.encode())
        self.x.stdin.flush()


    def __init__(self, speaker=0):
        """ Initiallize a festival speech synthesizer port.
        speaker: 0=British, 1=American, 2=American
        """
        # Test that Festival is installed
        which = os.popen('which festival').read()
        if which == '':
            self.installed = False
            print("Warning: Festival not found. Install festival for speech synthesis")
            return
        self.installed = True

        voice = ("voice_rab_diphone", "voice_don_diphone", "voice_ked_diphone")
        self.x = subprocess.Popen(['festival'], stdin=subprocess.PIPE)
        # rab, don, ked 
        cmd = ""

        self.comm("(audio_mode 'async)")
        # specifying a speaker give an err? "(%s)" % (voice[speaker]))
        self.comm(cmd)


    def hello_world(self):
        self.comm("""(SayText "Hello World!")""")


    def speak(self,text):
        """ Speak synchronously, works great for saying shorter segments """
        if not self.installed:
            return
        self.kill()
        text = text.replace("\"","'")
        self.comm("""(tts_text "%s" nil)""" % text)
        
        
    def kill(self):
        """ It turns out these "cancellation" features don't really help.
        This might work to break up speek_asyn, but we currently only
        use synchronous mode, rendering this useless.
        """
        if not self.installed:
            return
        self.comm("""(audio_mode 'shutup)""")
        #self.comm("""(audio_mode 'async)""")



class Festival_server:
    """ Festival Class
        This python class provides access to the Festival 
        speech synthesizer through its client/server capacity.

        It requires Festival to be running on the target machine.
 
          festival --server 
          festival --server --server_passwd ESKIMOS_ONLY
          festival --server server_access_list 

        On some machines, Festival_server is denied permission
        to connect to its self. This has not been diagnosed.
        If you want correct behavior you should use the
        Festival class, which relies on simple UNIX Pipes.
    """

    # telnet socket 
    ftel = None

    def __init__(self, machine_name="localhost", port=1314, passwd=None):
        try: 
          self.open(machine_name, port, passwd)
        except socket.error: 
          if machine_name != "localhost":
            raise "Could not connect to Festival Server"
          else:
            #if we couldnt telnet, try lanuching the service ourself!
            print("starting festival --server")
            
            errcode = subprocess.call("festival --server > /dev/null &", shell=True)
            if errcode == 0:
                print("Started Festival")
            else:
                raise "Failed to Start Festival"
            
        # TODO: replace sleep(1) with a real test
        # There *must* be a more reliable way to do this
        for i in range(5):
            try:
                self.open(machine_name,port,passwd)
                print("Started Festival")
                return
            except socket.error:
                pass
            time.sleep(1)
            
        print("Failed to Start Festival")

            
    def open(self, machine_name,port,passwd):
        """ Connect to a festival server """
        #print "telnetting to festival --server on " + \
        #    " %s port %d " % (machine_name, port)
        self.ftel = Telnet(machine_name, port)
        if passwd != None:
          self.ftel.write(passwd + "\n")
        time.sleep(.1)

        #self.command("voice_rab_diphone") # British Gordon (default)
        #self.command("voice_don_diphone") # American Donovan
        self.command("voice_ked_diphone") # American 
        time.sleep(.1)
        #self.command("Parameter.set 'Audio_Command ")
        #             """ "esdcat -m -r $SR $FILE" """)
        #time.sleep(.1)
        #self.command("Parameter.set 'Audio_Method 'Audio_Command")
        time.sleep(.1)
        self.command("audio_mode 'async")
        time.sleep(.1)
        

    def command(self, text):
        """Perform a command transaction with the festival server"""
        self.ftel.write("(" + text + ")\n")

    def write(self, text):
        """send scheme text to the festival server """
        self.ftel.write(text)
        
    def speak(self, text):
        """ Command the festival server to speak """
        self.kill()
        tmp = tempfile.mktemp()
        open(tmp,"w").write(text)
        self.command("""tts "%s" nil""" % tmp)

    def kill(self):
        self.command("audio_mode 'shutup")


festival_global = None
def say(text):
    global festival_global
    festival_global = festival_global if festival_global else Festival()
    festival_global.speak(text)


def play():
    f = Festival()
    #f.speak("hello")
    say("hello")
    print("Type some text and press return to hear it spoken: ")
    print("Type ^d or . to quit.")
    while True:
        try:
            txt = input(prompt)
        except EOFError:
            break
        if txt == '.':
            break
        f.speak(txt)


if __name__ == "__main__":
    if len(sys.argv) > 1:
        msg = ' '.join(sys.argv[1:])
        print(msg)
        say(msg)
    else:
        play()
