#!/usr/bin/env python3
# MorseCode.py
import time
import os
import sys
import logging
import re
from pprint import pprint
from pathlib import Path

try:
    from changelog_version import version
except:
    # When imported by setup.py we sub-import this way
    from src.changelog_version import version

# package version info
author  = "Pehr Anderson"
author_email = "pehr@pehr.net"
date = "2017-04-10"
url = "http://pehr.net/morseall"
license = "GPL"


def get_src_dir():
    "Returns the directory of the morseall source code"
    return Path(sys.argv[0]).parent.absolute()


class MorseCode:
    """This class has the following features:
    loads, parses, and maintains the configuration dictionary
    decodes morse code events into characters
    re-codes configuration dictionary into help pages
    """

    def __init__(self, mode=1):
        "Initialize Morse code processing with settings from conf."
        self.code = ""
        self.previous_code = ""    
        self.word = ""
        # number of clicks & releases
        self.count = 0 
        # step in milliseconds to inc/dec the dot_time_ms
        self.time_step = 25

        self.state = 0;
        self.next_event = None
        self.dur = None
        self.time = time.time()
        self.conf = ConfigFile()

        # read required parameters in conf
        self.update_conf()
        self.mode = mode


    def get_conf_dir(self):
        "Get the configuration directory"
        return self.conf.sound_dir.parent


    def update_conf(self):
        """Update the MorseCode object with the latest from ConfigFile
        Use defaults if values are not in the config file"""
        self.dot_time_ms    = self.conf.get_int("dot_time_ms", 250)
        self.char_time_ms   = self.conf.get_int("char_time_ms", self.dot_time_ms*4)
        self.word_time_ms   = self.conf.get_int("word_time_ms",  self.dot_time_ms*12) 
        self.repeat_time_ms = self.conf.get_int("repeat_time_ms", self.dot_time_ms*4)
        self.repeat_rate_ms = self.conf.get_int("repeat_rate_ms", self.dot_time_ms)
        self.repeat_enable  = self.conf.get_int("repeat_enable",  1)
        self.autospace_enable = self.conf.get_int("autospace_enable", 0)
        self.speak_enable   = self.conf.get_int("speak_enable", 1)
        self.sound_enable   = self.conf.get_int("sound_enable", 1)
        self.font_size      = self.conf.get_int("font_size", 14)
        self.rows = self.conf.get_int("rows",   24)
        self.columns        = self.conf.get_int("columns", 80)
        self.text_color     = self.conf.get_str("text_color", "black")


    def callback(self,event,count=None):
        """Morse Event Callback, to advance the state machine
        State Machine Description:
        1 - last event was button-down, it might be a dot
        2 - last event was dot-timeout, its at least a dash
        3 - last event was repeat-timeout, repeat the previous char
        3 - last event was repeat-continue, repeat chars at a high rate
        4 - last event was end-pressed, wait for repeat
        5 - last event was end-timeout, repeat previous char
        5 - last event was end-continue, repeat at a high rate
        10 - last event was button-up, there might be more to the char
        11 - last event was char-timeout, we finished a char
        0 - last event was word-timeout, we finished a whole word

        callback returns (code, , , )
        code -- when the code is ready to be parsed, it is here
        """
        get_int = self.conf.get_int

        #logging.debug('%s %s %s %s' % (self.state, self.code, event, count))
        text = None
        # by default there is no next_event
        self.next_event = None
        # duration to next_event in milliseconds
        self.dur = None
        
        def set_next_event(self, event, timeout):
            # Specify when the next event should be
            self.next_event = event
            self.dur = timeout
            self.time = time.time()

        # Button Callbacks
        if event == "pressed":
            # Clear our word if we are starting from rest
            if self.state == 0:
                self.word = ""
            self.state = 1
            set_next_event(self, "dot-timeout", self.dot_time_ms)
            # start by appending a dot
            self.code = self.code + "."
            # increment the count to invalidate scheduled timeout events
            self.count = self.count + 1
                
        elif event == "released":
            if self.state == 3:
                # do nothing if we have been repeating chars
                self.state = 0
                return (text, self.next_event, self.dur, self.count)
            # a redundant test for (possibly) better performance
            #if self.state == 1 and self.PercentComplete() == 1:
            #    self.code = self.code[:-1] + "-"
            self.state = 10
            set_next_event(self, "char-timeout", self.char_time_ms)
            self.count = self.count + 1

        elif event == "dash":
            # The user pressed the Dash button!
            # Clear our word if we are starting from rest
            if self.state == 0:
                self.word = ""
            # start by appending a dash
            self.code = self.code + "-"
            self.state = 10
            set_next_event(self, "char-timeout", self.char_time_ms)
            self.count = self.count + 1

        elif event == "end-pressed":
            # The user pressed the Char-End button!
            #if self.state in [0,11]:
            # if there is no active character, get one from the past!
            if self.code == "":
                self.code = self.previous_code
            # disable any existing timers
            self.count = self.count + 1
            # skip to the end of character state
            self.state = 4
            set_next_event(self, "end-timeout", self.word_time_ms)
            # parse our character 
            text = self.parse_code()

        elif event == "end-released":
            self.state = 11
            # disable any existing timers
            self.count = self.count + 1
            set_next_event(self, "word-timeout", self.word_time_ms)
            
        elif count != self.count:
            # timeouts are only valid if we haven"t clicked in between
            return (text, self.next_event, self.dur, self.count)
   
        # Timeout Callbacks
        elif self.state == 1 and event == "dot-timeout":
            self.state = 2
            # replace the appended dot with a dash
            self.code = self.code[:-1] + "-"
            # schedule an event to start repeating after a while
            if self.repeat_enable:
                set_next_event(self,"repeat-timeout", self.repeat_time_ms)
            
        elif self.state == 2 and event == "repeat-timeout":
            self.state = 3
            set_next_event(self, "repeat-continue", self.repeat_rate_ms)
            # delete the last dash when we are repeating
            self.code = self.code[:-1]
            # if there is there no code, repeat previous
            if self.code == "": 
                self.code = self.previous_code 
            text = self.parse_code()

        elif self.state == 3 and event == "repeat-continue":
            set_next_event(self, "repeat-continue", self.repeat_rate_ms)
            self.code = self.previous_code
            text = self.parse_code()

        elif self.state == 4 and event == "end-timeout":
            self.state = 5
            set_next_event(self, "end-continue", self.repeat_rate_ms)
            # if there is there no code, repeat previous
            if self.code == "": 
                self.code = self.previous_code 
            text = self.parse_code()

        elif self.state == 5 and event == "end-continue":
            set_next_event(self, "end-continue", self.repeat_rate_ms)
            self.code = self.previous_code
            text = self.parse_code()
            
        elif self.state == 10 and event == "char-timeout":
            self.state = 11
            set_next_event(self, "word-timeout", self.word_time_ms)
            text = self.parse_code()
            
        elif self.state == 11 and event == "word-timeout":
            self.state = 0
            if self.autospace_enable:
                x = self.previous_code
                self.code = self.conf.get_code("<space>")
                text = self.parse_code()
                # if we repeat, we don"t want to repeat the space
                self.previous_code = x

        # return code and any generated events
        if text:
            return (self.previous_code, self.next_event, self.dur, self.count)
        return ("", self.next_event, self.dur, self.count)


    def parse_code(self):
        """Parse a morse code characater, returns the decoded text
        update self.word and self.previous_code"""
        self.previous_code = self.code
        text = self.lookup()
        self.code = ""
        # update self.word, such that it acurately represents our state
        if text and len(text) > 0:
            c = text[0]
            if "A" <= c <= "z" or "0" <= c <= "9":
                self.word = self.word + text
            else:
                self.word = text
        else:
            self.word = text
        return text


    def function(self,text):
        """Execute a keystroke funtion, eg '<faster>' or '<slower>'. 
        Returns 1 if recognized, 0 for unknown"""
        if text == "<slower>":
            self.dot_time_ms = self.dot_time_ms + self.time_step
            self.repeat_rate_ms = self.dot_time_ms
            self.conf.set("dot_time_ms", self.dot_time_ms)
            self.conf.set("repeat_rate_ms", self.repeat_rate_ms)
        elif text == "<faster>":
            self.dot_time_ms = self.dot_time_ms - self.time_step
            if self.dot_time_ms < 25:
                self.dot_time_ms = 25
            self.repeat_rate_ms = self.dot_time_ms
            self.conf.set("dot_time_ms", self.dot_time_ms)
            self.conf.set("repeat_rate_ms", self.repeat_rate_ms)
        else:
            return 0
        return 1


    def get_state(self):
        "Get the string representing the state of the morse code engine"
        states = { 0:"rest", 1:"dot", 2:"dash", 3:"repeat", 4:"end", 5:"end-rpt", 10:"char", 11:"word"}
        return states[self.state]

 
    def PercentComplete(self):
        "Get the fraction of time left in this state from 0.999 down to 0.0"
        if self.dur == None:
            return 1
        return (time.time() - self.time) * 1000 / self.dur


    def get_duration(self):
        "Get the timeout duration of the current state in milliseconds"
        if self.dur == None:
            raise "No duration associated with this state"
        return self.dur


    def lookup(self):
        "Get the text generated by the current pattern in the morse code engine."
        try:
            return self.conf.get_text(self.code, self.mode)
        except:
            return ''


    def set_mode(self, mode):
        "Set the code page, an integer between 1 and max_modes"
        self.mode = mode
        

    def help_pages(self, rows=13, columns=60):
        """returns a list of strings, each string is 
        a block of text no bigger than rows x columns
        that contains key mapping and configuration info.
        """
        # set the spacing. Keys are right justified, Vals are left justified.
        pad_keyval = 1
        pad_column = 2

        # Figure out how wide the columns will be on the help page.
        maxlen = 0
        code_list = sorted(self.conf.items())
        for key,val in code_list:
            maxlen = max(len(key)+len(val), maxlen)

        # help_cols is the number of columns on each page
        help_cols = 1
        sum = maxlen + pad_keyval + pad_column
        while maxlen + pad_keyval + help_cols * sum <= columns: 
            help_cols = help_cols + 1

        # sort and organize the keymap and settings for viewing
        keymap = []
        settings = []
        for key,val in self.conf.items():
            if key[0] in ".-01234567890": 
                keymap.append((val,key))
            else:
                settings.append((key,val))
        # don't sort the key names, use the previously sorted .- codes
        #keymap.sort()
        settings.sort()

        # start stuffing the pages with data.
        def collapse(mat,pad,help_cols,columns):
            lines = []
            for r in range(len(mat)):
                line = (" "*pad).join(mat[r])
                mat[r] = [""]*help_cols 
                lines.append(line[:columns])
            return '\n'.join(lines)

        r, c, mat = 0, 0, []
        for i in range(rows):
            mat.append([""]*help_cols)
        pages = []
        for key,val in keymap + settings:
            len_pad = pad_keyval + maxlen - len(key) - len(val)
            mat[r][c] = key + " "*len_pad + val
            r = r + 1
            if r >= rows:
                r = 0
                c = c + 1
                if c >= help_cols:
                    # We have filled our page, so go on to the next!
                    pages.append(collapse(mat,pad_column,help_cols,columns))
                    c = 0
        # collapse the last page
        if r or c:
            pages.append(collapse(mat,pad_column,help_cols,columns))
        return pages
        """
        i = 0
        line = ""
        lines = []
        pages = []
        for key,val in keymap + settings:
            len_pad = pad_keyval + maxlen - len(key) - len(val)
            pad = " "*len_pad
            line = line + key + pad + val
            i=i+1
            # if we are at the end of a line, add it to the page. 
            if i % help_cols == 0:
                lines.append(line[:columns])
                line = ""
                # if we are at the end of a page, add it to pages[]
                if len(lines) >= rows: 
                    pages.append("\n".join(lines))
                    lines = []
                else:
                    line = line + (" "*pad_column)
        # don't forget to make the last page 
        if line or lines:
            pages.append("\n".join(lines)+"\n"+line[:columns])
        """
        return pages


    def set(self, code, val):
        "Set a variable in the current morseall engine"
        self.conf.set(code, val)
        self.update_conf()


    def get_int(self, code, default=None):
        return self.conf.get_int(code, default)


    def get_text(self, code):
        """takes a code, returns text dependent on current mode.
        If the code contains a mode, it overrides the current mode"""
        return self.conf.get_text(code, self.mode)
        

    def get_code(self, text):
        "takes text, returns a code dependent on current mode."
        return self.conf.get_code(text, self.mode)

    
    def get_code_soundfile(self, code):
        """takes a code, returns a soundfile dependent on current mode.
        If the code contains a mode, it overrides the current mode"""
        return self.conf.get_code_soundfile(code, self.mode)


    def get_text_soundfile(self, text):
        "takes text, returns a soundfile dependent on current mode."
        return self.conf.get_text_soundfile(text, self.mode)

    

class ConfigFile:
    """Configuration data for Morseall,
    Read, Mutate, and Hold

    The format is two or three columns ascii text
        variable-name   NUMBER
        ..code...       string     sound-file
        # comments prefixed by the pound sign
    """
    
    def __init__(self):
        "read configuration file(s) and return them as one string"
        self.conf  = {} 
        self.code2soundfile  = {}
        self.text2code = {}
        self.text2soundfile  = {}

        src_dir = get_src_dir()

        # look in several places for morseall.conf
        config_directories = [ 
            "/usr/share/morseall",
            "/usr/local/share/morseall",
            src_dir.parent / "morseall-data",
            src_dir / "morseall-data",
            "/etc",
            Path("~/.config/morseall").expanduser()
        ]
        sound_directories = [ 
            "/usr/share/morseall/sound",
            "/usr/local/share/morseall/sound",
            src_dir / "morseall-data" / "sound",
            src_dir.parent / "morseall-data" / "sound",
        ]
        # list of morseall.conf files that were found and read
        # read in the config data, later files overwrite settings
        def which_dirs_have_file(dirs, file):
            files = [Path(d) / file for d in dirs]
            conf_files = [f.parent for f in files if f.exists()]
            if conf_files:
                return conf_files
            msg = f"Could not find {file} in {dirs=}\n\n{files=}\n\n"
            logging.error(msg)
            print(msg)
            sys.exit(5)

        self.conf_files = which_dirs_have_file(config_directories, 'morseall.conf')
        self.conf_files = [f / 'morseall.conf' for f in self.conf_files]
        for f in self.conf_files:
            self.read_file(f)
        if self.conf.__len__() == 0:
            logging.error(f"Config files contained no data {self.conf_files}")

        # Set the sound_dir variable
        self.sound_dir = which_dirs_have_file(sound_directories, '0.opus')[0]


    def get_sound_dir(self):
        "Get the directory where the sound files are located"
        return self.sound_dir


    def read_file(self, filename):
        "Parse a configuration file, Stores the keys in self.conf"
        if not os.access(filename, os.O_RDONLY):
            return -1
        logging.debug("Reading config file %s" % filename)
        contents = open(filename).read()
        lines = contents.split("\n")
        for i in range(len(lines)):
            # strip comments and whitespace from a line
            uncommented_line = lines[i].split("#",1)[0]
            line = uncommented_line.strip()
            if not line:
                continue
            # parse the line with expected format "code text sounfile.opus"
            pattern = r"\s*([\w\-\.]+)\s+([\w\-\.\<\>]+)\s*([\w\.\-_]*)"
            matches = re.findall(pattern, line)
            if matches:
                self.add_entry(matches[0][0], matches[0][1], matches[0][2])
            else:
                logging.error('could not parse line %d in %s: %s' % (i+1, filename, line))
        return 0

    def add_entry(self, code, text, soundfile=''): 
        "Adds a code mapping to the configureation"
        if code in self.conf:
           if self.conf[code] != text:
                logging.info(f"Redefinig {code} from {self.conf[code]} to {text}")
        self.conf[code] = text
        mode = self.get_mode_from_modecode(code)
        modetext = '%d%s' % (mode,text)
        logging.debug('adding %-12s %-20s %s' % (code, text, soundfile))
        if code[0] in '0123456789':
            # for a non-default mode, set the text for the specific mode.
            self.text2code[modetext] = code
        else: # for the default mode, set the default text.
            self.text2code[text] = code
        if soundfile:
            self.code2soundfile[code] = soundfile
            self.text2soundfile[text] = soundfile

    def get_mode_from_modecode(self, modecode):
        "returns the mode prefix stripped from a morse code snippet, eg: '2-.-' returns 2"
        i = re.findall(r'^(\d+)',modecode)
        if i:
            return int(i[0])
        return 1

    def split_modecode(self, modecode):
        "returns int(mode), str(raw_code).  raw_code is the code with the mode prefix stripped off"
        match = re.findall(r'^(\d+)([\.\-]*)$',modecode)
        if match:
            return int(match[0]), match[1]
        return 1, modecode


    def get_text(self, code, mode):
        "Get the text that a bit of Morse code generates, returns <none> if no match"
        modecode = '%d%s' % (mode,code)
        if modecode in self.conf:
            return self.conf[modecode]
        elif code in self.conf:
            return self.conf[code]
        return None
    

    def get_code(self, text, mode):
        "Get the Morse code snippet for a given bit of text, returns ??? if no match"
        # look for a mode-specific mapping for this text
        modetext = '%d%s' % (mode,text)
        if modetext in self.text2code:
            return self.text2code[modetext]
        modetext = modetext.lower()
        if modetext in self.text2code:
            return self.text2code[modetext]
        # try the text with no mode
        if text in self.text2code:
            return self.text2code[text]
        text =text.lower()
        if text in self.text2code:
            return self.text2code[text]
        return None
    
        
    def get_int(self, key, default=None):
        "return a configuration variable as an integer, can take a default value"
        return int(self.conf.get(key,default))
    

    def get_str(self, key, default=None):
        "return a configuration variable as a string, can take a default value"
        return self.conf.get(key,default)
    

    def set(self, code, val):
        "Set a configuration value, always saves as a string."
        self.conf[code] = str(val)


    def items(self):
        "Return a list of configuration items [(key, value), ...]"
        return self.conf.items()


    def get_soundfile_from_code(self, code, mode):
        "Get the sound file associated with a code"
        modecode = '%d%s' % (mode, code)
        if modecode in self.code2soundfile:
            return self.code2soundfile[modecode]
        elif code in self.code2soundfile:
            return self.code2soundfile[code]
        return None


    def get_soundfile_from_text(self, text):
        "Get the sound file associated with a bit of text"
        # look for a mode-specific mapping for this text
        if text in self.text2soundfile:
            return self.text2soundfile[text]
        text = text.lower()
        if text in self.text2soundfile:
            return self.text2soundfile[text]
        return None
    

def test():
    mc = MorseCode()
    print(f'Version "{version}"')


if __name__ == '__main__':
    test()

