#!/usr/bin/env python3
import sys
import os
import time
import threading
import select
import collections
import re
import logging

import Xlib
import Xlib.display
import Xlib.ext
import Xlib.ext.xtest
import Xlib.Xcursorfont
from Xlib import X, Xutil


# These modifiers can be toggled by command
keycode_modifiers = ['shift','altgr','numlock']

# Map each modifier to the keyname that triggers it
modifier2keyname = {
    'shift': 'Shift_L',
    'control': 'Control_L',
    'altgr': 'Alt_L',
    'super': 'Super_L',
    'numlock': 'Num_Lock',
 }

keyname2modifier = {
    'Shift_L': 'shift',
    'Shift_R': 'shift',
	'Control_L': 'control',
	'Control_R': 'control',
    'Alt_L': 'altgr',
    'Alt_R': 'altgr',
	'Super_L': 'super',
	'Super_R': 'super',
	'Num_Lock': 'numlock',
}


class FakeXThread (threading.Thread):
    """
    Create a tread to stuff X extension commands into
    """
    
    def __init__(self, callback=None, display=None, verbose=False):
        """ Create a FakeXThread for stuffing X extension commands
        to control the mouse and keyboard.
        Takes one opitonal argument to specify an Xlib Display.
        If you don't know what that means, just call it with no arguments.
        """
        self.verbose = verbose

        # set the name of our thread
        threading.Thread.__init__(self, name="FakeXThread")
        self.grab_state = 0

        if display == None:
            display = Xlib.display.Display()
        self.d = display
        self.objects = []

        # clear the mouse callback
        self.mouse_callback = callback

        # Find which screen to open the window on
        self.screen = self.d.screen()
 
        self.window = self.screen.root.create_window(
            50, 50, 300, 200, 2, self.screen.root_depth,
            X.InputOutput,
            X.CopyFromParent,

            # special attribute values
            background_pixel = self.screen.white_pixel,
            event_mask = (
                X.ExposureMask |
                X.StructureNotifyMask |
                X.ButtonPressMask |
                X.ButtonReleaseMask |
                X.Button1MotionMask |
                X.Button2MotionMask |
                X.Button3MotionMask),
                colormap = X.CopyFromParent,
            )
     
        self.gc = self.window.create_gc(
        foreground = self.screen.black_pixel,
        background = self.screen.white_pixel,
        )

        # Set some WM info
        self.WM_DELETE_WINDOW = self.d.intern_atom('WM_DELETE_WINDOW')
        self.WM_PROTOCOLS = self.d.intern_atom('WM_PROTOCOLS')
        self.window.set_wm_name('Xlib example: draw.py')
        self.window.set_wm_icon_name('draw.py')
        self.window.set_wm_class('draw', 'XlibExample')

        self.window.set_wm_protocols([self.WM_DELETE_WINDOW])
        self.window.set_wm_hints(flags = Xutil.StateHint, initial_state = Xutil.NormalState)

        self.window.set_wm_normal_hints(
        flags = (Xutil.PPosition | Xutil.PSize
            | Xutil.PMinSize), min_width = 20, min_height = 20)

        # Map the window, making it visible
        self.window.map()
        self.load_keycodes()


    def load_keycodes(self):
        'Load the system keymaps from /usr/share/ibus/keymaps/'
        common = open('/usr/share/ibus/keymaps/common').read()
        us = open('/usr/share/ibus/keymaps/us').read()
        modifiers = open('/usr/share/ibus/keymaps/modifiers').read()
        self.keycode2name = collections.OrderedDict()
        self.name2keycode = collections.OrderedDict()
        text = common + us + modifiers
        pattern = r'\s*([a-z0-9 ]*) *keycode ([0-9]+) = ([a-zA-Z0-9_]+)[ addupper]*'
        #for line in text.split('\n'):
        #    modifier, keycode, name = re.findall(pattern, line)
        Key = collections.namedtuple('Key', ['modifier', 'keycode'])
        if self.verbose:
            print(f"FakeX Key Map:\n_name_\tmodifier\tkeycode")
        for modifier, keycode, name in re.findall(pattern, text):
            mk = Key(modifier, int(keycode))
            self.keycode2name[mk] = name
            self.name2keycode[name] = mk
            if self.verbose:
                print(f"{name}\t{modifier}\t{keycode}")


    def get_key_by_name(self, keyname):
        'Return the Key for a name: k.modifier and k.keycode'
        keycode = self.name2keycode.get(keyname, None)
        if not keycode:
            logging.error('Error: unknown keyname %s' % keyname)
        return keycode


    def fake_keyname(self, keyname):
        'Execute a fake keystroke by name including required modifiers'
        key = self.get_key_by_name(keyname)
        mod_keycodes = []
        for mod in key.modifier.split():
            mod_keycodes += [self.name2keycode.get(modifier2keyname[mod])]
        # press the modifiers, press the key, release key, etc.
        for c in mod_keycodes:
            self.fake_keycode_press(c)
        self.fake_keycode_press(key.keycode)
        self.fake_keycode_release(key.keycode)
        for c in mod_keycodes:
            self.fake_keycode_release(c)


    # Main loop, handling events
    def run(self, grab_state=0):
        #self.current_pos = None
        self.t = 0
        # handle pending X events
        i = self.d.pending_events()
        while i > 0:
            event = self.d.next_event()
            self.handle_x_event(event)
            i = i - 1

        # set grab state if we haven't already...
        if self.t == 0:
            self.grab_set(grab_state)
            self.t = 1
                
        while 1:
            timeout = 1  # set the timeout in seconds
            readable, w, e = select.select([self.d], [], [], timeout)
            # if no files are readable, it is a timeout
            if not readable:
                # handle timeouts here!
                #print("_timeout_")
                self.fake_keyname('t')
            
            # handle pending X events
            i = self.d.pending_events()
            while i > 0:
                event = self.d.next_event()
                self.handle_x_event(event)
                i = i - 1


    def set_mouse_callback(self, func):
        self.mouse_callback = func
        
            
    def handle_x_event(self, e):
        # Window has been destroyed, quit
        if e.type == X.DestroyNotify:
            self.exit(0)

        # Some part of the window has been exposed, 
        # redraw all the objects.
        if e.type == X.Expose:
            for o in self.objects:
                o.expose(e)

        # Left button pressed, start to draw
        if e.type == X.ButtonPress and e.detail == 1:
            print("down")
            #self.fake_keycode_press(0xFF96)
            #self.fake_keycode_release(0xFF96)
            self.fake_keyname('1')


        # Left button released, finish drawing
        if e.type == X.ButtonRelease and e.detail == 1:
            print("up")

        if e.type == X.ButtonPress and e.detail == 2:
            print("2down")
            self.fake_keycode_press(0xFF98)
            self.fake_keycode_release(0xFF98)


        if e.type == X.ClientMessage:
            if e.client_type == self.WM_PROTOCOLS:
                fmt, data = e.data
            if fmt == 32 and data[0] == self.WM_DELETE_WINDOW:
                self.exit()


    def exit(self):
        "perhaps we should alert the calling thread before we quit?"
        sys.exit(0)

        
    def fake_keysym(self, keysym):
        keycode = self.d.keysym_to_keycode(keysym)
        self.fake_keycode_press(keycode)
        self.fake_keycode_release(keycode)


    def fake_keycode_press(self, keycode):
        self.d.xtest_fake_input(X.KeyPress, detail=keycode)

        
    def fake_keycode_release(self, keycode):
        self.d.xtest_fake_input(X.KeyRelease, detail=keycode)
    

    def fake_mouse_move(self, x, y):
        """ move the mouse relative to the current positoin"""
        self.d.warp_pointer(x, y)
        # this would move the mouse to absolute coordinates
        #self.d.xtest_fake_input(X.MotionNotify, x=x, y=y)
        
        
    def fake_mouse_button(self, button):
        self.d.xtest_fake_input(X.ButtonPress, detail=1)
        self.d.xtest_fake_input(X.ButtonRelease, detail=1)


    def grab_set(self, enable):
        """
        owner_events
        Specifies a Boolean value that indicates whether
        keyboard events are to be reported as usual.
        False: all generated key events are reported
               with respect to grab_window
        True:  if a generated key event would normally be reported
               to this client, it is reported normally; otherwise the
        event is reported with respect to the grab_window.

        pointer_mode
        Specifies further processing of ointer events.
        You can pass GrabModeSync or GrabModeAsync.
        If pointer_mode is GrabModeAsync, pointer event processing
        is unaffected by activation of the grab.
        If pointer_mode is GrabModeSync, the state of the pointer
        (as seen by client applications) appears to freeze,
        and teh X server generates no further pointer events
        until the grabbing client issues a releasing XAllowEvents
        call or until the keyboard (sic?) grab is released.
        Actual pointer changes are not lost while the pointer
        is frozen; they are simply queued in the server for
        later processing.

        If the keyboard 
        """
        self.grab_state = 1 if enable else 0
        gtime = 0
        if enable == 0:
            return self.d.ungrab_pointer(gtime)
        
        owner_events = True
        event_mask = X.ButtonPressMask | X.ButtonReleaseMask
        pointer_mode = X.GrabModeAsync
        keyboard_mode = 0 # X.GrabModeAsync
        confine_to = 0
        cursor = 0   # Xlib.Xcursorfont.center_ptr
        self.window.grab_pointer(owner_events, event_mask,
                                 pointer_mode, keyboard_mode,
                                 confine_to, cursor, gtime)


if __name__ == '__main__':
    x = FakeXThread(verbose=True)
    x.start()
    print('Waiting for no reason')
    for i in range(9):
        keyname = f'{i}'
        key = x.get_key_by_name(keyname)
        print(f'"{keyname}" -- {key=}')
        x.grab_set = i%1
        print(i)
        if key:
            x.fake_keyname(keyname)
        # we could do anything here...
        time.sleep(1)
    print("end")
    
    
