#!/usr/bin/env python3
"""MorseAll - the morse code shell     http://pehr.net/morseall

#    Copyright (C) 2000-2025 Pehr Anderson <pehr@pehr.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""
import argparse
import logging
import os
import re
import sys
from urllib.parse import unquote

# Local imports 
import MorseCode
#import xmouser
#import fakex
from msound import MorseSound
#import evdev
#from evdev import ecodes as ec
import key_mouse
import changelog_version
from key_mouse import SHIFT, CTRL, ALT, UP, DOWN

# GUI Imports
try: 
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    from gi.repository import Gdk
    from gi.repository import GLib
    #from gi.repository import GObject
    gi.require_version('Pango', '1.0')
    from gi.repository import Pango
    gi.require_version('Vte', '2.91')
    from gi.repository import Vte
    gi.require_version('GSound', '1.0')
    from gi.repository import GSound
except ImportError: 
    print('You need to install some packages:')
    print('As of Ubuntu 24.04, GSound bindings are in gir1.2-gsound-1.0')
    print('  sudo apt install python3-gi festival gir1.2-gsound-1.0 python3-evdev')
    print('Make yourself a member of the "input" group to allow grabbing.')
    print('  sudo usermod -a -G input $USER')
    sys.exit(3)


class MorseallGnome:
    "Main Application"

    def __init__(self, morse, args):
        """ Initialize the Gui, including all graphic elements
        Requires a morse object and a configuration object.
        # TODO, make MorseGnome able to handle resizing
        """
        self.grab_state = 0   # current user-controllable grab setting
        self.mode_line = ''
        self.help_line = ''
        self.help_pagenum = 0
        self.morse = morse
        self.args = args

        self.key_mouse = key_mouse.Key_Mouse(mouse_callback = self.event_callback)

        # start building a GUI already!
        self.win = Gtk.Window()
        self.win.set_title("morseall-" + MorseCode.version)
        self.win.connect("delete-event", Gtk.main_quit)
        self.win.connect("leave_notify_event", self.leave_window_callback)
        self.win.set_resizable(False)

        #vertical box, vte on top, a grid of labels on bottom
        vbox = Gtk.VBox()
        self.win.add(vbox)
        vbox.show()

        self.init_term(vbox, morse.rows, morse.columns)
        self.init_canvas(vbox)
        self.resize(morse.columns, morse.rows)
        self.poll()
        self.win.show()

        self.win.connect("button_press_event",   self.event_callback)
        self.win.connect("button_release_event", self.event_callback)
        self.win.connect("key_press_event",   self.event_callback)
        # For some reason we have to bind VTE clics separately
        self.term.connect("button_press_event",   self.event_callback)
        self.term.connect("button_release_event", self.event_callback)
        self.term.connect("key_press_event",   self.event_callback)

        # ???
        #master_fd, stdin = self.pty.get_inputs()
        #Gtk.input_add(self.master_fd, Gdk.INPUT_READ, self.pty_callback)

        # forward stdin into the terminal
        ###GObject.io_add_watch(sys.__stdin__, Gdk.INPUT_READ, self.pty_callback)

        # take mouse events away from vte widget to prevent selections
        self.win.grab_add()

        # Before we show, let's set our Modes proper...
        self.toggle_mode(morse.mode)
        self.sound_dev = MorseSound(morse)

        # show_all must be called before pointer_grab will work
        self.win.show_all()
        self.grab_flush()
        # handle the command line argument -g/--grab to override default
        if self.args.grab:
            self.toggle_grab(1)


    def pty_callback(self, fn, x):
        "Stuff characters into the vte terminal widget for display"
        try:
            text = os.read(fn, 1024)
            self.term.feed(text, len(text))
        except OSError:
            self.pty_launch()
        return True


    def pty_launch(self):
        "Launch the shell process to be run in the Vte PseudoTerminal"

        # If the shell dies, we want it to restart using our custom bashrc file
        fn = str((self.morse.get_conf_dir() / "bashrc").absolute())
        cmd = ["/bin/bash", "-c", f"while true; do /bin/bash -rcfile {fn}; done"]
        # Deprecated since 0.48:  self.term.spawn_sync().  Use self.term.spawn_async()
        #Vte_terminal_spawn_async(VteTerminal*, 
            #VtePtyFlags, 
            #const char*, 
            #char**, 
            #char**, 
            #GSpawnFlags, 
            #GSpawnChildSetupFunc, 
            #gpointer, 
            #GPid*, 
            #GCancellable*, 
            #GError**)
        # For a VTE Example see: https://gitlab.gnome.org/GNOME/vte/-/issues/221
        pty = Vte.Pty.new_sync(Vte.PtyFlags.DEFAULT)
        gs = GLib.SpawnFlags
        spawn_flags = gs.DEFAULT 
        #spawn_flags |= gs.CHILD_INHERITS_STDIN 
        #spawn_flags |= gs.DO_NOT_REAP_CHILD
        pty.spawn_async(
            None,
            cmd,
            [], # envv Add more environment variables at launch
            spawn_flags,
            None,
            None,
            -1, 
            None,
            self.ready
            )
        self.term.set_pty(pty)
        '''# Create a box with scrollers
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        scroller = Gtk.ScrolledWindow()
        scroller.set_hexpand(True)
        scroller.set_vexpand(True)
        scroller.add(self.term)
        box.pack_start(scroller, False, True, 2)
        self.win.add(box)
        '''

    def ready(self, pty, task):
        'TODO: Does this Vte callback needs to do anything?'
        print(f"ready()")# {pty=} {task=}")


    def init_term(self, vbox, rows, columns):
        "Initialize the terminal, only needs to be done once"
        self.term = Vte.Terminal()
        # TODO modify_font is deprecated, so we need a CSS Style to replace it?
        #self.term.modify_font(self.get_font())
        self.term.set_scrollback_lines(49)
        #def child_died_event(object):
        #    sys.exit(0)
        #term.connect("child_died", child_died_event)

        # Scroll-on output is crucial if we have a scroll bar since
        # users could trap themselves unable to see their text with
        # a slip of the mouse.
        self.term.set_scroll_on_output(1)
        vbox.pack_start(self.term, True, True, 0)
        self.pty_launch()
        self.term.show()

        # signal handler to set window hints for character-accurate resizing
        #w = self.term.charwidth
        #h = self.term.charheight
        #self.win.set_geometry_hints(geometry_widget=term,
        #       min_width=2*w, min_height=2*h,
        #       base_width=w,  base_height=h,
        #       width_inc=w,   height_inc=h)


    def resize(self,columns,rows):
        "Resize the text terminal to a specific number of characters"
        logging.debug('screen columns,rows = %d,%d' % (columns,rows))
        self.term.set_size(columns,rows)


    def init_canvas(self, vbox):
        "Setup labels to show morse-code states and help messages"
        #h = self.term.charheight
        h = 20
   
        # add the bottom half, set its size and put it in the vbox widget
        self.grid = grid = Gtk.Grid() #spacing = 1)
        grid.set_column_spacing(30)
        grid.set_row_spacing(0)
        grid.set_size_request(700, 155)
        vbox.pack_start(grid, True, True, 0)

        # create morse assistants (left column)
        #font = self.get_font()
        self.label_state = Gtk.Label()
        # TODO modify_font is deprecated, so we need a CSS Style to replace it?
        #self.label_state.modify_font(font)
        grid.attach(self.label_state, 1, 0, 1, 5)

        # create toggle widgets (center column)
        gc = self.morse.get_code
        self.tog_shift = MyToggle("Shift   "+gc("Shift_L"), tristate=1, key=SHIFT)
        self.tog_ctrl  = MyToggle("Control "+gc("Control_L"), tristate=1, key=CTRL)
        self.tog_alt   = MyToggle("Alt     "+gc("Alt_L"), tristate=1, key=ALT)
        self.tog_sound = MyToggle("Sound   "+gc("<sound>"))
        self.tog_grab  = MyToggle("Grab    "+gc("<grab>"))

        grid.attach(self.tog_shift.label, 0, 0, 1, 1)
        grid.attach(self.tog_ctrl.label,  0, 1, 1, 1)
        grid.attach(self.tog_alt.label,   0, 2, 1, 1)
        grid.attach(self.tog_sound.label, 0, 3, 1, 1)
        grid.attach(self.tog_grab.label,  0, 4, 1, 1)

        # create help widget (right column)
        self.label_help = Gtk.Label()
        # TODO modify_font is deprecated, so we need a CSS Style to replace it?
        #self.label_help.modify_font(font)
        self.label_help.set_justify(Gtk.Justification.CENTER)
        grid.attach(self.label_help,  2, 0, 1, 8)

        # set up toggle state
        self.tog_sound.set(self.morse.conf.get_int("sound_enable", 1))

        # start with the page that tells how to toggle grab_enable grab
        self.help_next_page("grab")

        # these toggles get their initial value from the config file
        self.tog_grab.set(self.morse.get_int("grab_enable",0))


    def get_font(self):
        font = self.morse.conf.get_str('font', default='courier')
        size = self.morse.conf.get_int('font_size', default=16)
        return Pango.FontDescription('%s %s' % (font, size))
        

    def poll(self):
        "Every tick of the morse clock, update these widgets"
        # make the updated text for our label
        # show only the ends of some strings if they get too long.
        lines =["code: %s" % self.morse.code[-15:],
            "time: %s" % self.morse.get_state(),
            "char: %s" % self.morse.lookup(),  
            "word: %s" % str(self.morse.word)[-20:],
            self.help_line,
            self.mode_line]
        self.label_state.set_text('\n'.join(lines))


    def morse_callback(self, event, counter=None):
        """Consolidated callback for all morse code events
        This callback will register its self for future timeout events
        """
        (code, event, msec, cnt) = self.morse.callback(event, counter)
        if code:
            # send keys to the terminal
            logging.debug('morse_callback code=%s' % code)
            text = self.morse.get_text(code)
            # save the mode in case it changes
            mode = self.morse.mode
            self.parse(text)
            if self.tog_sound.get():
                logging.info('Play the mode-specific version of this code %s', text)
                #self.sound_dev.play_text(text, mode)
                self.sound_dev.play_code(code, mode)
            else:
                logging.info('tog_sound not set')
        if event:
            #GObject.timeout_add(msec, self.morse_callback, event, cnt)
            GLib.timeout_add(msec, self.morse_callback, event, cnt)
        self.poll()


    def event_callback(self, widget, event=None):
        "Generic Event Callback, handles all possible Gdk events"
        if event.type == Gdk.EventType.BUTTON_PRESS:
             #self.win.emit_stop_by_name("button_press_event")
            if event.button == 1:
                self.morse_callback("pressed")
                return True
            if event.button == 2:
                self.morse_callback("end-pressed")
                return True
            if event.button == 3:
                self.morse_callback("dash")
                return True
        elif event.type == Gdk.EventType.BUTTON_RELEASE:
            #self.win.emit_stop_by_name("button_release_event")
            if event.button == 1:
                self.morse_callback("released")
                return True
            if event.button == 2:
                self.morse_callback("end-released")
                return True
        elif event.type == Gdk.EventType.KEY_PRESS:
            self.keyval_callback(event)
            return True
        return False



    def keyval_callback(self, event):
        """Process a keyboard event from the real physical keyboard.
        Real keypresses should work as well as morse code sequences.
        This makes morseall easier to try out & learn.
        Pressing the real Control, Shift, or Alt will activate the Toggles.
        """
        try:
            #this line segfaults for some unknown reason
            #c = Gdk.keyval_name(event.keyval)
            c = key_mouse.keyval2char[event.keyval]
            self.parse(c)
        except KeyError:    
            if event.string:
                self.parse(event.string)
            else:
                logging.warn("Unknown keyval 0x%x" % (event.keyval))

            
    def leave_window_callback(self, widget, event=None):
        "Triggered whenever the mouse attempts to leave the window"
        self.grab_flush()


    def toggle_mode(self, mode=None):
        "Advance the Code-handling mode - or set it if a mode is specified"
        if mode == None:
            mode = self.morse.mode + 1
        if mode > self.morse.conf.get_int("max_modes", 3):
            mode = 1 
        logging.debug('toggle_mode(%s) old=%s new=%s'% (mode,self.morse.mode,mode))
        self.morse.set_mode(mode)
        self.mode_line = "mode: %d %s" % (self.morse.mode, self.morse.get_code("<mode>"))
        self.help_next_page()
        return mode


    def toggle_grab(self, x=None):
        "Toggle Grab on and off, or set a value if one is given. 1=on 0=off"
        if x == None:
            x = self.morse.conf.get_int("grab_enable", 0)
            x = [1,0][x]
        if x not in [0,1]: 
            logging.warn('invalid grab value %d, should be in [0,1] % x')
        self.tog_grab.set(x)
        self.morse.conf.set("grab_enable", x)
        self.help_next_page("grab")
        self.grab_flush()
        return x


    def grab_flush(self):
        "Grab events from the whole screen"
        self.key_mouse.grab_set(self.grab_state)


    def parse(self, text):
        "Send keystrokes to the terminal"
        logging.debug('parse:  %s' % text)

        if text == "<speak>":
            self.speak()
        # toggles
        elif text == "Shift_L":
            self.tog_shift.toggle()
        elif text == "Control_L":
            self.tog_ctrl.toggle()
        elif text == "Alt_L":
            self.tog_alt.toggle()
            pass
        elif text == "<sound>":
            # read sound_enable state from ConfigFile object
            self.tog_sound.toggle()
            x = self.tog_sound.get()
            #self.morse.conf.set("sound_enable", x)
            logging.debug("sound.toggle = %d" % x)
            self.help_next_page("sound")
        elif text == "<grab>":
            x = self.toggle_grab()
            logging.debug("grab.toggle = %d" % x)
   
        # special functions
        elif text == "<slower>":
            self.morse.function(text)
            self.help_next_page("dot_time_ms")
            logging.debug("slower - dot_time_ms = %d" % self.morse.dot_time_ms)
        elif text == "<faster>":
            self.morse.function(text)
            self.help_next_page("dot_time_ms")
            logging.debug("faster - dot_time_ms = %d" % self.morse.dot_time_ms)
        elif text == "<reset>":
            self.pty_launch()
        elif text == "<help>":
            self.help_next_page()
        elif text == "<mode>":
            self.toggle_mode()
        elif re.findall(r'^<mode_(\d+)>$', text):
            m = re.findall(r'^<mode_(\d+)>$',text)[0].atoi()
            self.toggle_mode(m)
        elif text == "<exit>":
            Gtk.main_quit()
        else:
            self.handle_keypress(text)


    def handle_keypress(self, text):
        'For a PHYSICAL keystoke, instert it into VTE directly'
        #if kx:
        #    self.xthread.fake_keyname(text)
        #if keysym:
        #    self.xthread.fake_keysym(keysym)
        #elif keycode:
        #    self.xthread.fake_keycode_press(keycode)
        #    self.xthread.fake_keycode_release(keycode)
        #else:
        #    print('unkown how to handle')

        quoted = key_mouse.char2str.get(text)
        if quoted:
            text = unquote(quoted)
        else:
            # handle sticky characters
            if self.tog_shift.get():
                text = text.upper()
                # only cancel the shift if it isn't locked
                if self.tog_shift.get() == 1:
                    self.tog_shift.set(0)
            if self.tog_ctrl.get():
                text = chr(ord(text[0]) & 0x1f)
                # only cancel the control if it isn't locked
                if self.tog_ctrl.get() == 1:
                    self.tog_ctrl.set(0)
            if self.tog_alt.get():
                # alt has no effect on the terminal...
                if self.tog_alt.get() == 1:
                    self.tog_alt.set(0)
        self.send_to_term(text)

 
    def send_to_term(self, text):
        'Send ascii text characters to the terminal'
        # Turn unicode text into uint(8) bytes
        self.term.feed_child(text.encode())


    def speak(self):
        'Send all the text on the screen to speech engine'
        r = self.morse.rows
        c = self.morse.columns
        txt = self.term.get_text()
        self.sound_dev.say_text(txt)
        #try:
        #    Festival.say(txt)
        #except: 
        #    logging.info("Speech Engine Error: ?")


    def help_next_page(self, search_text=None):
        'Change to the next help page,  If search_text: go to the relevant page'
        # Initialize all the help pages based on screen size
        pages = self.morse.help_pages(6,43)

        # if we were given search_text, hunt for it!
        if search_text:
            for i in range(len(pages)):
                if re.search(re.escape(search_text), pages[i]):
                    page = i
                    break
        else:
            # Advance pagenum to the next page
            page = max(0, self.help_pagenum + 1)
            if page >= len(pages):
                page = 0

        # Set the help text to the new help page
        self.label_help.set_text(pages[page])
        self.help_line = "help(%2d/%2d) %s" % (page+1, len(pages), 
            self.morse.get_code("<help>"))
        self.help_pagenum = page


class MyToggle:

    def __init__(self, offtext, ontext=None, tristate=0, key=None):
        """MyToggle is a widget for tracking and displaying toggle state
        offtext is displayed when the toggle is off (default)
        ontext is displayed when the toggle is on
        If key is set, we set the key_mouse state.
        """
        self.state = 0
        self.key = key
        self.label = Gtk.Label()
        #Deprecated since 3.0:  self.label.modify_font(font), Use overrid_font.
        #Deprected since 3.16: self.label.override_font(font), Use a custom CSS style via GtkStyleProvider and CSS style class
        # ?? TODO:  How do I use CSS in Gtk!??
        self.label.set_justify(Gtk.Justification.LEFT)
        self.offtext = offtext
        self.ontext = ontext
        self.tristate = tristate
        if self.ontext == None: 
            self.ontext = self.offtext
        self.set(0)


    def set(self, state):
        "Set the toggle state to one of 0,1,2"
        assert state in [0,1] or (self.tristate and state in [0,1,2]), f'Invalid state {state=}'
        self.state = state
        stateChar = [" ","x","X"][state]
        self.label.set_text("[%c] %s"%(stateChar, self.ontext))
        # If this toggle has a keycode, give key_mouse the new value (UP or DOWN)
        #if self.key:
        #    key_mouse.set_keycode(self.key, [UP, DOWN, DOWN][state])


    def toggle(self):
        "toggle the widget, returns the current state"
        if self.tristate and self.state == 1:
            self.set(2)
        else:
            self.set(0 if self.state else 1)
        print(f'toggle {self.offtext=} {self.state=}')
        return self.state


    def get(self):
        "returns current state"
        return self.state


def parse_args():
    p = argparse.ArgumentParser(prog='morseall', 
        description = 'Morse Code Engine')
    ad = p.add_argument
    #ad('-t','--test', action='store_true',
    #   help='test every code in the configuration file')
    #ad('-c','--config', action='store', default=None,
    #   help='specify a configuration file')
    ad('-g','--grab', action='store_true',
        help='Grab control of the mouse (very useful for disabled users)')
    ad('-v','--verbose', action='store_true',
         help='show debugging information')
    ad('-V','--version', action='store_true')
    return p.parse_args()


def main():
    # create the necessary application objects
    args = parse_args()
    # handle version arguments
    if args.version:
       print(f"Morseall Version {MorseCode.version}")
       sys.exit(0)
    # set the logging verbosity level
    level = logging.WARN
    if args.verbose:
        level=logging.DEBUG
        print("set verbose")
    logging.basicConfig(level=level, datefmt="%Y-%m-%d %H:%M",
        format="%(asctime)s %(levelname)-8s %(message)s")
    # start morseall
    logging.debug("Starting Morseall")
    morse = MorseCode.MorseCode()
    logging.debug(f"Config Files:\n\t"+("\n\t".join([str(f) for f in morse.conf.conf_files])))
    logging.debug(f"sound_dir {morse.conf.sound_dir}")
    morseall = MorseallGnome(morse, args)
    # play the 'welcome to morseall' sound
    morseall.sound_dev.play_text("<reset>")
    # there is only one gnome Gtk.main. Let"s run it!
    try:
        Gtk.main()
    except KeyboardInterrupt as why:
        logging.error("Keyboard Interrupt: %s" % why)
        sys.exit()


if __name__ == "__main__":
    main()
