#!/usr/bin/env python3
"""MorseAll - the morse code shell     http://pehr.net/morseall
#    Copyright (C) 2000-2017 Pehr Anderson <pehr@pehr.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""
import logging
import re
import sys
import time
import gi
gi.require_version('GSound', '1.0')
from gi.repository import GSound
import Festival


class MorseSound:
    "Use the Gnome2 sound interface"

    def __init__(self, morse):
        "initialize the sound system, set the default path"
        self.morse = morse
        self.ctx = GSound.Context()
        self.ctx.init()
        self.sound_dir = self.morse.conf.get_sound_dir()

        
    def play_text (self, text, mode = 0):
        "Plays a soundfile or uses festival if soundfile is missing"
        logging.debug('Playing text(%s)' % text)
        soundfile = self.morse.conf.get_soundfile_from_text(text)
        if soundfile:
            f = self.sound_dir / soundfile
            if f.is_file():
                logging.info('play_text sound file (%s)' % f)
                self.play_file(f)
            else:
                logging.info('missing sound file (%s)' % soundfile)
                self.say_text(text)
            return
        logging.info('No soundfile for text(%s)' % text)
        self.say_text(text)


    def play_code(self, code, mode):
        "Plays a soundfile or uses festival if soundfile is missing"
        logging.debug('Playing code/mode(%s,%d)' % (code, mode))
        soundfile = self.morse.conf.get_soundfile_from_code(code, mode)
        if soundfile:
            f = self.sound_dir / soundfile
            if f.is_file():
                logging.info('play_code sound file (%s)' % f)
                self.play_file(f)
            else:
                logging.info('missing sound file (%s)' % soundfile)
                self.say_code(code)
            return
        logging.info('No soundfile for code, mode(%s, %d)' % (code, mode))
        self.say_text("<none>")


    def play_file(self, soundfile):
        "Plays a sound file, requires the full path to the file"
        self.ctx.play_simple({"media.filename": str(soundfile)})


    def say_code(self, code):
        "Convert a code into speakable text, and say it"
        text = self.morse.get_text(code)
        self.say_text(text)


    def say_text(self, text):
        "Convert character text into speakable text, and say it"
        d = {
        "<reset>"       :"Welcome to Morse All!",
        "a"             :"ay",
        "<help>"        :"help",
        "<sound>"       :"sound enable",
        "Shift_L>"      :"shift",
        "Alt_L"         :"alt",
        "Control_L"     :"control",
        "parenright"    :"close paren",
        "parenleft"     :"open paren",
        "braceleft"     :"open brace",
        "braceright"    :"close brace",
        "bracketleft"   :"open bracket",
        "bracketright"  :"close bracket",
        "quotedbl"      :"double quote",
        "KP_Right"      :"right",
        "KP_Up"         :"up",
        "KP_Down"       :"down",
        "KP_Left"       :"left",
        "Pointer_Right" :"",
        "Pointer_Up"    :"",
        "Pointer_Left"  :"",
        "Pointer_Down"  :"",
        "Pointer_Button1"  :"click",
        "Pointer_Button2"  :"click",
        "Pointer_Button3"  :"click",
        "Pointer_DblClick1" :"ckclick",
        "Pointer_DblClick2" :"ckclick",
        "Pointer_DblClick3" :"ckclick",
        "Pointer_Hold1" :"hold left",
        "Pointer_Hold2" :"hold middle",
        "Pointer_Hold3" :"hold right",
        }
        if text in d: 
            Festival.say(d[text])
            return
        # Strip unsayable characters from code names
        text = re.sub('^<', '', text)
        text = re.sub('>$', '', text)
        text = re.sub('_', ' ', text)
        text = re.sub('-', ' ', text)
        Festival.say(text)


    def play_all(self):
        'Play all the sounds'
        print(f'Playing all the sound files in {self.sound_dir}')
        print('Play the first sounds 1 second apart, and 3% faster each time.')
        seconds = 1
        for f in sorted(self.sound_dir.glob('*.opus')):
            self.play_file(f)
            time.sleep(seconds)
            seconds *= .97
        time.sleep(.5)


if __name__ == '__main__':
    import MorseCode
    morse = MorseCode.MorseCode()
    m_snd = MorseSound(morse=morse)
    m_snd.play_all()
